﻿using ArgumentValidation;
using System;
using System.Text;

namespace FrequencyDistribution
{
    public class WaveData
    {
        #region Format specification

        public int ChunkLength { get; private set; }

        public int ChannelsCount { get; private set; }

        public int SampleRate { get; private set; }

        public int AverageBytesPerSecond { get; private set; }

        public short BlockAlign { get; private set; }

        public short BitsPerSample { get; private set; }

        #endregion

        #region Samples

        public int SamplesPerChannel { get; private set; }

        public short[] SamplesLeftChannel { get; private set; }

        public short[] SamplesRightChannel { get; private set; }

        #endregion

        #region Constants

        private const string riffChunk = "RIFF";
        private const string waveChunk = "WAVE";
        private const string fmtChunk = "fmt ";
        private const string dataChunk = "data";

        private const int supportedBitsPerSample = 16;
        private const int fmtExtendedSize = 18;
        private const int minLength = 44;
        private const int chunkIdLength = 4;
        private const int pcmTag = 1;

        #endregion

        #region Fields

        private int shortSize = sizeof(short);
        private int intSize = sizeof(int);

        #endregion        

        public WaveData(byte[] rawData)
        {
            Check.IsNull(rawData);
            Check.NotLessThan(rawData.Length, minLength);

            int offset = ReadRiffChunk(rawData);

            offset = ReadFormatChunk(rawData, offset);

            ReadDataChunk(rawData, offset);
        }

        private void VerifyChunkId(byte[] rawData, int chunkIndex, string expectedValue)
        {
            var chunkId = Encoding.UTF8.GetString(rawData, chunkIndex, chunkIdLength);

            if (string.Compare(chunkId, expectedValue) != 0)
            {
                throw new ArgumentException("Incorrect chunk");
            }
        }

        private int ReadRiffChunk(byte[] rawData)
        {
            int offset = 0;

            VerifyChunkId(rawData, offset, riffChunk);
            offset += chunkIdLength;

            ChunkLength = BitConverter.ToInt32(rawData, offset);
            offset += intSize;

            VerifyChunkId(rawData, offset, waveChunk);
            offset += chunkIdLength;

            return offset;
        }

        private int ReadFormatChunk(byte[] rawData, int offset)
        {
            VerifyChunkId(rawData, offset, fmtChunk);
            offset += chunkIdLength;

            var formatChunkLength = BitConverter.ToInt32(rawData, offset);
            offset += intSize;

            var formatTag = BitConverter.ToInt16(rawData, offset);
            VerifyFormatTag(formatTag);
            offset += shortSize;

            ChannelsCount = BitConverter.ToInt16(rawData, offset);
            offset += shortSize;

            SampleRate = BitConverter.ToInt32(rawData, offset);
            offset += intSize;

            AverageBytesPerSecond = BitConverter.ToInt32(rawData, offset);
            offset += intSize;

            BlockAlign = BitConverter.ToInt16(rawData, offset);
            offset += shortSize;

            BitsPerSample = BitConverter.ToInt16(rawData, offset);
            VerifyBps();
            offset += shortSize;

            if (formatChunkLength == fmtExtendedSize)
            {
                var extensionLength = BitConverter.ToInt16(rawData, offset);
                offset += extensionLength + shortSize;
            }

            return offset;
        }

        private void VerifyFormatTag(int formatTag)
        {
            if (formatTag != pcmTag)
            {
                throw new ArgumentException("Unsupported data format");
            }
        }

        private void VerifyBps()
        {
            if (BitsPerSample != supportedBitsPerSample)
            {
                throw new ArgumentException("Unsupported sample bit depth");
            }
        }

        private void ReadDataChunk(byte[] rawData, int offset)
        {
            VerifyChunkId(rawData, offset, dataChunk);
            offset += chunkIdLength;

            ReadAudioData(rawData, offset);
        }

        private void ReadAudioData(byte[] rawData, int offset)
        {
            var dataSize = BitConverter.ToInt32(rawData, offset);
            var samplesIndex = offset + sizeof(int);

            SamplesPerChannel = dataSize / (sizeof(short) * ChannelsCount);

            SamplesLeftChannel = new short[SamplesPerChannel];
            SamplesRightChannel = new short[SamplesPerChannel];

            for (int i = 0, j = samplesIndex; i < SamplesPerChannel; i++, j += ChannelsCount * sizeof(short))
            {
                SamplesLeftChannel[i] = BitConverter.ToInt16(rawData, j);

                if (ChannelsCount > 1)
                {
                    SamplesRightChannel[i] = BitConverter.ToInt16(rawData, j + sizeof(short));
                }
                else
                {
                    SamplesRightChannel[i] = SamplesLeftChannel[i];
                }
            }
        }
    }
}
